/**
 * @internal
 */
const CONTROLLER = '__FUIDT_CONTROLLER__';

/**
 * @internal
 */
const ELEMENT_METADATA = '__FUIDT_ELEMENT_METADATA__';

/**
 * @internal
 */
const HTML_ELEMENT_REFERENCE = '__FUIDT_HTML_ELEMENT_REFERENCE__';

const createController = () => {
  let selectedElement = null;
  const controller = {
    get selectedElement() {
      return selectedElement;
    },
    select: nextSelectedElement => {
      if (isHTMLElementWithMetadata(nextSelectedElement)) {
        selectedElement = nextSelectedElement;
        observer.observe(nextSelectedElement.parentElement, {
          childList: true,
          subtree: false
        });
      }
      if (selectedElement && nextSelectedElement) {
        if (isNextSelectedElementValid(selectedElement, nextSelectedElement)) {
          return selectedElement;
        }
      }
      controller.withdraw();
      return selectedElement;
    },
    withdraw: () => {
      selectedElement = null;
      observer.disconnect();
    }
  };
  const observer = new MutationObserver(mutations => {
    if (!selectedElement) {
      return;
    }
    for (const mutation of mutations) {
      if (mutation.type === 'childList' && Array.from(mutation.removedNodes).includes(selectedElement)) {
        controller.withdraw();
      }
    }
  });
  return controller;
};
const injectController = targetDocument => {
  if (!targetDocument.defaultView) {
    return;
  }
  if (!targetDocument.defaultView[CONTROLLER]) {
    targetDocument.defaultView[CONTROLLER] = createController();
  }
};
const isNextSelectedElementValid = (selectedElement, nextSelectedElement) => {
  const metadata = selectedElement[ELEMENT_METADATA];
  switch (metadata.type) {
    case 'middleware':
      return metadata.references.has(nextSelectedElement);
    default:
      return false;
  }
};
const isHTMLElementWithMetadata = element => Boolean(element && ELEMENT_METADATA in element && element.parentElement !== null);

/**
 * Verifies if a given node is an HTMLElement,
 * this method works seamlessly with frames and elements from different documents
 *
 * This is preferred over simply using `instanceof`.
 * Since `instanceof` might be problematic while operating with [multiple realms](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/instanceof#instanceof_and_multiple_realms)
 *
 * @example
 * ```ts
 * isHTMLElement(event.target) && event.target.focus()
 * isHTMLElement(event.target, {constructorName: 'HTMLInputElement'}) && event.target.value // some value
 * ```
 *
 */
function isHTMLElement(element, options) {
  var _typedElement$ownerDo, _options$constructorN;
  const typedElement = element;
  return Boolean((typedElement == null || (_typedElement$ownerDo = typedElement.ownerDocument) == null ? void 0 : _typedElement$ownerDo.defaultView) && typedElement instanceof typedElement.ownerDocument.defaultView[(_options$constructorN = options == null ? void 0 : options.constructorName) != null ? _options$constructorN : 'HTMLElement']);
}

/**
 * @internal
 */

let counter = 0;
const generateReferenceId = () => {
  return HTML_ELEMENT_REFERENCE + ":" + counter++;
};
const createReferences = () => {
  const weakSet = new WeakSet();
  const map = new Map();
  const references = {
    add: element => {
      const id = generateReferenceId();
      map.set(id, element);
      weakSet.add(element);
      return id;
    },
    get: id => {
      const element = map.get(id);
      if (element && weakSet.has(element)) {
        return element;
      }
    },
    has: element => {
      return weakSet.has(element);
    }
  };
  return references;
};

const serialize = data => {
  const references = createReferences();
  const serializedData = JSON.parse(JSON.stringify(data, (_, value) => {
    if (isHTMLElement(value)) {
      return references.add(value);
    }
    if (typeof value === 'object' && value && Object.getPrototypeOf(value) !== Object.prototype && Object.getPrototypeOf(value) !== Array.prototype) {
      if ('toString' in value) {
        return value.toString();
      }
      return undefined;
    }
    return value;
  }));
  return [serializedData, references];
};

/**
 * devtools middleware
 * @public
 */
const devtools = function (targetDocument, middlewareDataCallback) {
  if (middlewareDataCallback === void 0) {
    middlewareDataCallback = floatingUIMiddlewareDataCallback;
  }
  return {
    name: '@floating-ui/devtools',
    fn: state => {
      injectController(targetDocument);
      const [serializedData, references] = serialize(middlewareDataCallback(state));
      Object.assign(state.elements.floating, {
        [ELEMENT_METADATA]: {
          references,
          serializedData,
          type: 'middleware'
        }
      });
      return {};
    }
  };
};
const floatingUIMiddlewareDataCallback = state => ({
  ...state,
  type: 'FloatingUIMiddleware'
});

export { devtools, devtools as middleware };
